// Copyright (c) 2006 dean gaudet <dean@arctic.org>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.

#include <inttypes.h>
#include "vtime64.h"
#include "rdtsc.h"

#define NS_SCALE (10)

// note the zero parameter wouldn't be in a production implementation... it's
// here only to ensure the computation always returns cpu 0.  this is
// appropriate for testing -- and means i don't depend on the actual value of
// the various registers.  i just pass it in so that gcc is fooled and doesn't
// optimize things away.

static inline ulong vgetcpu_sgdt(ulong zero)
{
  struct {
    char pad[6];
    uint16_t limit;
    uint64_t base;
  } gdt;
  asm volatile("sgdt %0" : "=m" (gdt.limit));
  return (0x10000 - gdt.limit) & zero;
}

static inline ulong vgetcpu_sldt(ulong zero)
{
  ulong nr;
  asm volatile("sldt %0" : "=r" (nr));
  // XXX: i probably need (nr >> 1) & 0x7f or something here
  return nr & zero;
}

static inline ulong vgetcpu_sidt(ulong zero)
{
  struct {
    char pad[6];
    uint16_t limit;
    uint64_t base;
  } idt;
  asm volatile("sidt %0" : "=m" (idt.limit));
  return (idt.limit & 0xfff) & zero;
}

static inline ulong vgetcpu_lsl(ulong zero)
{
  ulong nr;
  asm volatile("lsl %1,%0" : "=r" (nr) : "r" (0ul));
  return nr & zero;
}

#define unlikely(x)  __builtin_expect(!!(x), 0)

// we need to stop gcc from hoisting the restart test... but
// we don't need a true cpu memory barrier (because any case
// where the code was rescheduled would have to restart the
// instructions anyhow).
#define gcc_rmb() do {asm volatile("":::"memory");} while (0)

// fake per-cpu data structure
struct {
  uint64_t tsc_base;
  uint64_t tsc_scale;
  uint64_t epoch;
} per_cpu[32];

#define vtime64_template(name)                                  \
  uint64_t vtime64_##name(ulong zero)                           \
  {                                                             \
    uint64_t result;                                            \
    ulong cpu_nr;                                               \
    uint64_t tsc_base;                                          \
    do {                                                        \
      cpu_nr = vgetcpu_##name(zero);                            \
      tsc_base = per_cpu[cpu_nr].tsc_base;                      \
      uint64_t now = rdtsc() - tsc_base;                        \
      result = (now * per_cpu[cpu_nr].tsc_scale) >> NS_SCALE;   \
      result += per_cpu[cpu_nr].epoch;                          \
      gcc_rmb();                                                \
    } while (unlikely(per_cpu[cpu_nr].tsc_base != tsc_base      \
                      || cpu_nr != vgetcpu_##name(zero)));      \
    return result;                                              \
  }

vtime64_template(sgdt)
vtime64_template(sldt)
vtime64_template(sidt)
vtime64_template(lsl)


// the following are a challenge to use... they rely on the kernel
// detecting that it is about to restart a task whose pc is pointing
// at the vsyscall page and then restarting the vsyscall (which is
// arbitrarily complex given the vsyscall may have a stack -- easier
// if the vsyscalls are implemented in assembly).  it's presented
// for interest sake because the timings are just that much better
// perhaps someone wants to go to the complete effort.

#define vtime64_restart_template(name)                          \
  uint64_t vtime64_restart_##name(ulong zero)                   \
  {                                                             \
    uint64_t result, now;                                       \
    ulong cpu_nr;                                               \
    cpu_nr = vgetcpu_##name(zero);                              \
    now = rdtsc() - per_cpu[cpu_nr].tsc_base;                   \
    result = (now * per_cpu[cpu_nr].tsc_scale) >> NS_SCALE;     \
    result += per_cpu[cpu_nr].epoch;                            \
    return result;                                              \
  }

vtime64_restart_template(sgdt)
vtime64_restart_template(sldt)
vtime64_restart_template(sidt)
vtime64_restart_template(lsl)
